@@ -111,6 +111,7 @@ struct segment {
 struct bitlk_signature {
 	uint8_t boot_code[3];
 	uint8_t signature[8];
+	uint16_t sector_size;
 } __attribute__ ((packed));
 
 struct bitlk_superblock {
@@ -414,6 +415,7 @@ void BITLK_bitlk_fvek_free(struct bitlk_fvek *fvek)
 		return;
 
 	crypt_free_volume_key(fvek->vk);
+	free(fvek);
 }
 
 void BITLK_bitlk_vmk_free(struct bitlk_vmk *vmk)
@@ -503,6 +505,18 @@ int BITLK_read_sb(struct crypt_device *cd, struct bitlk_metadata *params)
 		goto out;
 	}
 
+	params->sector_size = le16_to_cpu(sig.sector_size);
+	if (params->sector_size == 0) {
+		log_dbg(cd, "Got sector size 0, assuming 512.");
+		params->sector_size = SECTOR_SIZE;
+	}
+
+	if (!(params->sector_size == 512 || params->sector_size == 4096)) {
+		log_err(cd, _("Unsupported sector size %" PRIu16 "."), params->sector_size);
+		r = -EINVAL;
+		goto out;
+	}
+
 	/* read GUID and FVE metadata offsets */
 	if (read_lseek_blockwise(devfd, device_block_size(cd, device),
 		device_alignment(device), &sb, sizeof(sb), fve_offset) != sizeof(sb)) {
@@ -721,6 +735,7 @@ int BITLK_dump(struct crypt_device *cd, struct device *device, struct bitlk_meta
 	log_std(cd, "Info for BITLK%s device %s.\n", params->togo ? " To Go" : "", device_path(device));
 	log_std(cd, "Version:      \t%u\n", params->metadata_version);
 	log_std(cd, "GUID:         \t%s\n", params->guid);
+	log_std(cd, "Sector size:  \t%u [bytes]\n", params->sector_size);
 	log_std(cd, "Created:      \t%s", ctime((time_t *)&(params->creation_time)));
 	log_std(cd, "Description:  \t%s\n", params->description);
 	log_std(cd, "Cipher name:  \t%s\n", params->cipher);
@@ -898,7 +913,7 @@ static int decrypt_key(struct crypt_device *cd,
 {
 	char *outbuf;
 	int r;
-	uint32_t key_size = 0;
+	uint16_t key_size = 0;
 
 	outbuf = crypt_safe_alloc(enc_key->keylength);
 	if (!outbuf)
@@ -913,10 +928,12 @@ static int decrypt_key(struct crypt_device *cd,
 	}
 
 	/* key_data has it's size as part of the metadata */
-	memcpy(&key_size, outbuf, 4);
-	key_size = le32_to_cpu(key_size);
+	memcpy(&key_size, outbuf, 2);
+	key_size = le16_to_cpu(key_size);
 	if (enc_key->keylength != key_size) {
-		log_err(cd, _("Wrong key size."));
+		log_err(cd, _("Unexpected key data size."));
+		log_dbg(cd, "Expected key data size: %zu, got %" PRIu16 "", enc_key->keylength, key_size);
+
 		r = -EINVAL;
 		goto out;
 	}
@@ -1155,6 +1172,9 @@ int BITLK_activate(struct crypt_device *cd,
 		}
 	}
 
+	if (params->sector_size != SECTOR_SIZE)
+		dmd.flags |= CRYPT_ACTIVATE_IV_LARGE_SECTORS;
+
 	r = dm_targets_allocate(&dmd.segment, num_segments);
 	if (r)
 		goto out;
@@ -1175,7 +1195,7 @@ int BITLK_activate(struct crypt_device *cd,
 						segments[i].iv_offset,
 						segments[i].iv_offset,
 						NULL, 0,
-						SECTOR_SIZE);
+						params->sector_size);
 		if (r)
 			goto out;
 
